################################################################################
#
# Copyright (c) 2009 The MadGraph5_aMC@NLO Development team and Contributors
#
# This file is a part of the MadGraph5_aMC@NLO project, an application which 
# automatically generates Feynman diagrams and matrix elements for arbitrary
# high-energy processes in the Standard Model and beyond.
#
# It is subject to the MadGraph5_aMC@NLO license which should accompany this 
# distribution.
#
# For more information, visit madgraph.phys.ucl.ac.be and amcatnlo.web.cern.ch
#
################################################################################
"""Unit test Library for importing and restricting model"""
from __future__ import division

import copy
import os
import sys
import time

import tests.unit_tests as unittest
import madgraph.core.base_objects as base_objects
import models.import_ufo as import_ufo
import models.model_reader as model_reader
import madgraph.iolibs.export_v4 as export_v4
import models as ufomodels
import madgraph.various.misc as misc

_file_path = os.path.split(os.path.dirname(os.path.realpath(__file__)))[0]


#===============================================================================
# TestImportUFO
#===============================================================================
class TestImportUFO(unittest.TestCase):
    """Test class for the RestrictModel object"""

    def setUp(self):
        """Set up decay model"""
        #Read the full SM
        sm_path = import_ufo.find_ufo_path('heft')
        self.base_model = import_ufo.import_full_model(sm_path)

    def test_coupling_hierarchy(self):
        """Test that the coupling_hierarchy is set"""
        self.assertEqual(self.base_model.get('order_hierarchy'),
                         {'QCD': 1, 'QED': 2, 'HIG':2, 'HIW': 2})
         
    def test_expansion_order(self):
        """Test that the expansion_order is set"""
        self.assertEqual(self.base_model.get('expansion_order'),
                         {'QCD': 99, 'QED': 99, 'HIG':1, 'HIW': 1})

class TestNFlav(unittest.TestCase):
    """Test class for the get_nflav function"""

    def test_get_nflav_sm(self):
        """Tests the get_nflav_function for the full SM.
        Here b and c quark are massive"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_full_model(sm_path)
        self.assertEqual(model.get_nflav(), 3)

    def test_get_nflav_sm_nobmass(self):
        """Tests the get_nflav_function for the SM, with the no-b-mass restriction"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_model(sm_path + '-no_b_mass')
        self.assertEqual(model.get_nflav(), 5)

    def test_get_nflav_sm_nomasses(self):
        """Tests the get_nflav_function for the SM, with the no_masses restriction"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_model(sm_path + '-no_masses')
        self.assertEqual(model.get_nflav(), 5)

class TestGetQuarkPDG(unittest.TestCase):
    """Test class for the get_nflav function"""

    def test_get_quark_pdgs_sm(self):
        """Tests the get_quark_pdg_function for the full SM.
        Here b and c quark are massive"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_full_model(sm_path)
        self.assertEqual(model.get_quark_pdgs(), [-3, -2, -1, 1, 2, 3])

    def test_get_quark_pdgs_sm_nobmass(self):
        """Tests the get_quark_pdg_function for the SM, with the no-b-mass restriction"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_model(sm_path + '-no_b_mass')
        self.assertEqual(model.get_quark_pdgs(), [-5, -4, -3, -2, -1, 1, 2, 3, 4, 5])

    def test_get_quark_pdgs_sm_nomasses(self):
        """Tests the get_quark_pdg_function for the SM, with the no_masses restriction"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_model(sm_path + '-no_masses')
        self.assertEqual(model.get_quark_pdgs(), [-5, -4, -3, -2, -1, 1, 2, 3, 4, 5])

class TestNLeps(unittest.TestCase):
    """Test class for the get_nflav function"""

    def test_get_nleps_sm(self):
        """Tests the get_nleps_function for the full SM.
        Here all leptons have a mass"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_full_model(sm_path)
        self.assertEqual(model.get_nleps(), 0)

    def test_get_nleps_sm_nobmass(self):
        """Tests the get_nleps_function for the SM, with the no-b-mass restriction
        here the electron and muon are massless"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_model(sm_path + '-no_b_mass')
        self.assertEqual(model.get_nleps(), 2)

    def test_get_nleps_sm_nomasses(self):
        """Tests the get_nleps_function for the SM, with the no_masses restriction
        here the three leptons are massless"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_model(sm_path + '-no_masses')
        self.assertEqual(model.get_nleps(), 3)

class TestGetLuarkPDG(unittest.TestCase):
    """Test class for the get_nflav function"""

    def test_get_lepton_pdgs_sm(self):
        """Tests the get_lepton_pdg_function for the full SM.
        Here all leptons have a mass"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_full_model(sm_path)
        self.assertEqual(model.get_lepton_pdgs(), [])

    def test_get_lepton_pdgs_sm_nobmass(self):
        """Tests the get_lepton_pdg_function for the SM, with the no-b-mass restriction
        here the electron and muon are massless"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_model(sm_path + '-no_b_mass')
        self.assertEqual(model.get_lepton_pdgs(), [-13, -11, 11, 13])

    def test_get_lepton_pdgs_sm_nomasses(self):
        """Tests the get_lepton_pdg_function for the SM, with the no_masses restriction
        here the three leptons are massless"""
        sm_path = import_ufo.find_ufo_path('sm')
        model = import_ufo.import_model(sm_path + '-no_masses')
        self.assertEqual(model.get_lepton_pdgs(), [-15, -13, -11, 11, 13, 15])

class TestImportUFONoSideEffect(unittest.TestCase):
    """Test class for the the possible side effect on a UFO model loaded when
       converting it to a MG5 model"""

    def test_ImportUFONoSideEffectLO(self):
        """Checks that there are no side effects of the import of the LO UFO sm"""       
        ufo_model = ufomodels.load_model(import_ufo.find_ufo_path('sm'),False)
        original_all_particles = copy.copy(ufo_model.all_particles)
        original_all_vertices = copy.copy(ufo_model.all_vertices)
        original_all_couplings = copy.copy(ufo_model.all_couplings)
        original_all_lorentz = copy.copy(ufo_model.all_lorentz)
        original_all_parameters = copy.copy(ufo_model.all_parameters)
        original_all_orders = copy.copy(ufo_model.all_orders)
        original_all_functions = copy.copy(ufo_model.all_functions)

        ufo2mg5_converter = import_ufo.UFOMG5Converter(ufo_model)
        model = ufo2mg5_converter.load_model()
        # It is important to run import_ufo.OrganizeModelExpression(ufo_model).main() 
        # since this reverts some of the changes done in load_model()
        # There *is* side effects in-between, namely the expression of the CTcouplings
        # which contained CTparameters have been substituted to dictionaries.
        parameters, couplings = import_ufo.OrganizeModelExpression(ufo_model).main()        

        self.assertEqual(original_all_particles,ufo_model.all_particles)        
        self.assertEqual(original_all_vertices,ufo_model.all_vertices)
        self.assertEqual(original_all_couplings,ufo_model.all_couplings)
        self.assertEqual(original_all_lorentz,ufo_model.all_lorentz)
        self.assertEqual(original_all_parameters,ufo_model.all_parameters)
        self.assertEqual(original_all_orders,ufo_model.all_orders)
        self.assertEqual(original_all_functions,ufo_model.all_functions)

    def test_ImportUFOcheckgoldstone(self):
        """Check goldstone is correct in NLO UFO"""
        ufo_model = ufomodels.load_model(import_ufo.find_ufo_path('loop_qcd_qed_sm'),False)
        original_all_particles = copy.copy(ufo_model.all_particles)
        for part in original_all_particles:
            if part.name.lower() in ['g0','g+']:
                if hasattr(part,"GoldstoneBoson"):
                    self.assertEqual(part.GoldstoneBoson,True)
                elif hasattr(part,"goldstoneboson"):
                    self.assertEqual(part.goldstoneboson,True)
                else:
                    raise import_ufo.UFOImportError, "Goldstone %s has no attribute of goldstnoneboson in loop_qcd_qed_sm"%part.name
                    
        
    def test_ImportUFONoSideEffectNLO(self):
        """Checks that there are no side effects of the import of the NLO UFO sm"""
        ufo_model = ufomodels.load_model(import_ufo.find_ufo_path('loop_sm'),False)
        original_all_particles = copy.copy(ufo_model.all_particles)
        original_all_vertices = copy.copy(ufo_model.all_vertices)
        original_all_couplings = copy.copy(ufo_model.all_couplings)
        original_all_lorentz = copy.copy(ufo_model.all_lorentz)
        original_all_parameters = copy.copy(ufo_model.all_parameters)
        original_all_orders = copy.copy(ufo_model.all_orders)
        original_all_functions = copy.copy(ufo_model.all_functions)
        original_all_CTvertices = copy.copy(ufo_model.all_CTvertices)
        original_all_CTparameters = copy.copy(ufo_model.all_CTparameters)


        ufo2mg5_converter = import_ufo.UFOMG5Converter(ufo_model)
        model = ufo2mg5_converter.load_model()
        # It is important to run import_ufo.OrganizeModelExpression(ufo_model).main() 
        # since this reverts some of the changes done in load_model()
        # There *is* side effects in-between, namely the expression of the CTcouplings
        # which contained CTparameters have been substituted to dictionaries.
        parameters, couplings = import_ufo.OrganizeModelExpression(ufo_model).main()        

        self.assertEqual(original_all_particles,ufo_model.all_particles)
        self.assertEqual(original_all_vertices,ufo_model.all_vertices)
        self.assertEqual(original_all_couplings,ufo_model.all_couplings)
        self.assertEqual(original_all_lorentz,ufo_model.all_lorentz)
        self.assertEqual(original_all_parameters,ufo_model.all_parameters)
        self.assertEqual(original_all_orders,ufo_model.all_orders)
        self.assertEqual(original_all_functions,ufo_model.all_functions)
        self.assertEqual(original_all_CTvertices,ufo_model.all_CTvertices)
        self.assertEqual(original_all_CTparameters,ufo_model.all_CTparameters)

        # Also test that one new lorentz struture has been added within the model
        # and that the associate optimization is working as expected.
        self.assertEqual(len(original_all_lorentz) + 1, len(model['lorentz']))
        new_l = [l for l  in model['lorentz'] if l not in original_all_lorentz][0]
        new_name = new_l.name
        self.assertEqual(new_l.name, 'R2RGA_VVVV1')
        # find interactions with that lorentz structure
        int_with_it = []
        for id, vertices in model.get('interaction_dict').items():
            if new_name in vertices['lorentz']:
                int_with_it.append(vertices)
        self.assertEqual(len(int_with_it), 2)
        # check the first one
        vert = int_with_it[0]
        pdg = [p['pdg_code'] for p in vert['particles']]
        self.assertEqual(pdg, [21,21,21,21])
        # check the equivalent vertex in the original model
        old_vert = [ v for v in ufo_model.all_CTvertices if pdg == [p.pdg_code for p in v.particles]]
        #pick one
        old_vert = old_vert[0]
        
        # find the number of coupling associate to this lorentz structure
        ind = vert['lorentz'].index(new_name)
        coup_name = [ c for ((l,col),c) in vert['couplings'].items() if l ==ind]
        nb_old = len([ c for ((l,col),c) in vert['couplings'].items() if c == coup_name[0]])
        nb_new = len([ c for ((l,col,k),c) in old_vert.couplings.items() if c.name == coup_name[0]])
        self.assertEqual(3*nb_old, nb_new)

#===============================================================================
# TestRestrictModel
#===============================================================================
class TestRestrictModel(unittest.TestCase):
    """Test class for the RestrictModel object"""

    def setUp(self):
        """Set up decay model"""
        #Read the full SM
        sm_path = import_ufo.find_ufo_path('sm')
        self.base_model = import_ufo.import_full_model(sm_path)

        model = copy.deepcopy(self.base_model)
        self.model = import_ufo.RestrictModel(model)
        self.restrict_file = os.path.join(_file_path, os.path.pardir,
                                     'input_files', 'restrict_sm.dat')
        self.model.set_parameters_and_couplings(self.restrict_file)
         
        
    def test_detect_special_parameters(self):
        """ check that detect zero parameters works"""        
        
        expected = set(['I3x32', 'etaWS', 'conjg__CKM3x2', 'CKM1x2', 'WT', 'I1x32', 'I1x33', 'I1x31', 'I2x32', 'CKM3x1', 'I2x13', 'I2x12', 'I3x23', 'I3x22', 'I3x21', 'conjg__CKM2x1', 'lamWS', 'conjg__CKM2x3', 'I2x23', 'AWS', 'CKM1x3', 'conjg__CKM3x1', 'I4x23', 'ymc', 'ymb', 'yme', 'CKM3x2', 'CKM2x3', 'CKM2x1', 'ymm', 'conjg__CKM1x3', 'Me', 'ym', 'I2x22', 'WTau', 'lamWS__exp__2', 'lamWS__exp__3', 'yc', 'yb', 'ye', 'MC', 'MB', 'MM', 'conjg__CKM1x2', 'I3x31', 'rhoWS', 'I4x33', 'I4x13'])
        zero, one = self.model.detect_special_parameters()
        result = set(zero)
        self.assertEqual(len(result), len(expected))

        self.assertEqual(expected, result)
        
        expected = set(['conjg__CKM3x3', 'conjg__CKM2x2', 'CKM1x1', 'CKM2x2', 'CKM3x3', 'conjg__CKM1x1'])
        result = set(one)
        self.assertEqual(expected, result)

        
        
    def test_detect_identical_parameters(self):
        """ check that we detect correctly identical parameter """
        
        expected=set([('MZ','MH')])
        result = self.model.detect_identical_parameters()
        result = [tuple([obj[0].name for obj in obj_list]) for obj_list in result]
        
        self.assertEqual(expected, set(result))
        
    def test_merge_identical_parameters(self):
        """check that we treat correctly the identical parameters"""
        
        parameters = self.model.detect_identical_parameters()
        self.model.merge_iden_parameters(parameters[0])
        
        
        #check that both MZ and MH are not anymore in the external_parameter
        keeped = '1*%s' % parameters[0][0][0].name
        removed = parameters[0][1][0].name
        for dep,data in self.model['parameters'].items():
            if dep == ('external'):
                for param in data:
                    self.assertNotEqual(param.name, removed)
            elif dep == ():
                found=0      
                for param in data:
                    if removed == param.name:
                        found += 1
                        self.assertEqual(param.expr, keeped)
                self.assertEqual(found, 1)
        
        # checked that the mass (and the width) of those particles identical
        self.assertEqual(self.model['particle_dict'][23]['mass'],
                         self.model['particle_dict'][25]['mass'])
        self.assertNotEqual(self.model['particle_dict'][23]['width'],
                         self.model['particle_dict'][25]['width'])
        

        
    def test_detect_zero_iden_couplings(self):
        """ check that detect zero couplings works"""
        
        zero, iden = self.model.detect_identical_couplings()
        
        # check what is the zero coupling
        expected = set(['GC_17', 'GC_16', 'GC_15', 'GC_14', 'GC_13', 'GC_19', 'GC_18', 'GC_22', 'GC_30', 'GC_20', 'GC_89', 'GC_88', 'GC_101', 'GC_102', 'GC_103', 'GC_42', 'GC_106', 'GC_107', 'GC_82', 'GC_43', 'GC_84', 'GC_85', 'GC_86', 'GC_105', 'GC_28', 'GC_29', 'GC_48', 'GC_44', 'GC_23', 'GC_46', 'GC_47', 'GC_26', 'GC_24', 'GC_25', 'GC_83', 'GC_87', 'GC_93', 'GC_92', 'GC_91', 'GC_90'])
        result = set(zero)
        self.assertEqual(len(expected), len(result))
        for name in result:
            self.assertEqual(self.model['coupling_dict'][name], 0)
        
        self.assertEqual(expected, result)        
        
        # check what are the identical coupling
        expected = [[('GC_100',1), ('GC_108',1), ('GC_49',1), ('GC_45',1), ('GC_40',1), ('GC_41',1), ('GC_104',1)],
                    [('GC_21', 1), ('GC_27', -1)],
                    [('GC_3', 1), ('GC_4', -1)],
                    [('GC_39', 1), ('GC_38', -1)],
                    [('GC_51', 1), ('GC_50', -1)],
                    #[('GC_53', 1), ('GC_52', -1)], #GC_52 is not assigned to a vertex to they are consider as different coupling order and not merged... not relevant anyway
                    [('GC_56', 1), ('GC_54', -1)],
                    [('GC_66', 1), ('GC_67', -1)],
                    [('GC_7', 1), ('GC_9', -1)],
                    [('GC_70', 1), ('GC_73', -1)],
                    [('GC_75', 1), ('GC_74', -1)],
                    [('GC_76', 1), ('GC_79', -1)],
                    [('GC_77', 1), ('GC_78', -1)],
                    [('GC_97', 1), ('GC_96', -1)]]
        expected.sort()
        iden.sort()
        self.assertEqual(expected, iden)

    def test_locate_couplings(self):
        """ check the creation of the coupling to vertex dict """
        
        for candidate in self.model['interactions']:
            if [p['pdg_code'] for p in candidate['particles']] == [5, 5, 25]:
                input_bbh = candidate
                coupling_bbh = candidate['couplings'][(0,0)]
            if [p['pdg_code'] for p in candidate['particles']] == [23, 23, 25, 25]:
                input_zzhh = candidate
                coupling_zzhh = candidate['couplings'][(0,0)]
            if [p['pdg_code'] for p in candidate['particles']] == [11, 12, 24]:
                input_wen = candidate
                coupling_wen = candidate['couplings'][(0,0)]
            if [p['pdg_code'] for p in candidate['particles']] == [22, 24, 24]:
                input_aww = candidate
                coupling_aww = candidate['couplings'][(0,0)]            
        
        
        target = [coupling_bbh, coupling_zzhh, coupling_wen, coupling_aww]
        sol = {coupling_bbh: [input_bbh['id']],
               coupling_zzhh: [input_zzhh['id']],
               coupling_wen: [43, 44, 45, 66, 67, 68],
               coupling_aww: [input_aww['id']]}
        # b b~ h // z z h h //w- e+ ve // a w+ w-
        
        self.model.locate_coupling()
        for coup in target:
            self.assertTrue(coup in self.model.coupling_pos)
            self.assertEqual(sol[coup], [v['id'] for v in self.model.coupling_pos[coup]])

  
    def test_merge_iden_couplings(self):
        """ check that the merged couplings are treated correctly:
             suppression and replacement in the vertex """
        
        self.model.locate_coupling()
        zero, iden = self.model.detect_identical_couplings()
        self.assertEqual(len(iden), 13)
        
        # Check that All the code/model is the one intended for this test
        target = [i for i in iden if len(i)==7][0] 
        target2 = [i[0] for i in target]
        GC = target2[0]
        
        check_content = [['d', 'u', 'w+'], ['s', 'c', 'w+'], ['b', 't', 'w+'], ['u', 'd', 'w+'], ['c', 's', 'w+'], ['t', 'b', 'w+'], ['e-', 've', 'w+'], ['m-', 'vm', 'w+'], ['tt-', 'vt', 'w+'], ['ve', 'e-', 'w+'], ['vm', 'm-', 'w+'], ['vt', 'tt-', 'w+']]
        content =  [[p.get('name') for p in v.get('particles')] \
               for v in self.model.get('interactions') \
               if any([c in target2 for c in v['couplings'].values()])]

        self.assertEqual(len(check_content),len(content))#, 'test not up-to-date'      

        vertex_id = [v.get('id') \
               for v in self.model.get('interactions') \
               if any([c in target2 for c in v['couplings'].values()])]


        for id in vertex_id:
            is_in_target = False
            for coup in self.model.get_interaction(id)['couplings'].values():
                if coup in target2:
                    is_in_target = True
            assert is_in_target == True, 'test not up-to-date'
        
        # check now that everything is fine
        self.model.merge_iden_couplings(target)
        for id in vertex_id:
            has_GC = False
            for coup in self.model.get_interaction(id)['couplings'].values():
                self.assertFalse(coup in target[1:])
                if coup == GC:
                    has_GC = True
            self.assertTrue(has_GC, True)
        
        # check that the same occur with opposite sign coupling
        target = [i for i in iden if len(i)==2][1] 
        target2 = [i[0] for i in target]
        GC = target2[0]
        
        check_content = [['a', 'w+', 'w+'], ['e-', 'e-', 'a'], ['mu-', 'mu-', 'a'], ['ta-', 'ta-', 'a']]
        content =  [[p.get('name') for p in v.get('particles')] \
               for v in self.model.get('interactions') \
               if any([c in target2 for c in v['couplings'].values()])]
        #content =  [[v.get('couplings').values() for p in v.get('particles')] \
        #       for v in self.model.get('interactions')]
        self.assertEqual(len(check_content),len(content))#, 'test not up-to-date'
        
        vertex_id = [v.get('id') \
               for v in self.model.get('interactions') \
               if any([c in target2[1:] for c in v['couplings'].values()])]
        
        for id in vertex_id:
            is_in_target = False
            for coup in self.model.get_interaction(id)['couplings'].values():
                if coup in target2:
                    is_in_target = True
            assert is_in_target == True, 'test not up-to-date'
        
        self.model.merge_iden_couplings(target)
        for id in vertex_id:
            has_GC = False
            for coup in self.model.get_interaction(id)['couplings'].values():
                self.assertFalse(coup in target[1:])
                if coup == '-%s' % GC:
                    has_GC = True
            self.assertTrue(has_GC, True)
        
                 

    def test_remove_couplings(self):
        """ check that the detection of irrelevant interactions works """
        
        for candidate in self.model['interactions']:
            if [p['pdg_code'] for p in candidate['particles']] == [5, 5, 25]:
                input_bbh = candidate
                coupling_bbh = candidate['couplings'][(0,0)]
            if [p['pdg_code'] for p in candidate['particles']] == [21, 21, 21, 21]:
                input_4g = candidate
                coupling_4g = candidate['couplings'][(0,0)]
        
        found_bbh = 0
        found_4g = 0
        for dep,data in self.model['couplings'].items():
            for param in data:
                if param.name == coupling_bbh: found_bbh +=1
                elif param.name == coupling_4g: found_4g +=1
        self.assertTrue(found_bbh>0)
        self.assertTrue(found_4g>0)
        
        # make the real test
        result = self.model.remove_couplings([coupling_bbh,coupling_4g])
        
        for dep,data in self.model['couplings'].items():
            for param in data:
                self.assertFalse(param.name in  [coupling_bbh, coupling_4g])

             
    def test_remove_interactions(self):
        """ check that the detection of irrelevant interactions works """
        
        for candidate in self.model['interactions']:
            if [p['pdg_code'] for p in candidate['particles']] == [5, 5, 25]:
                input_bbh = candidate
                coupling_bbh = candidate['couplings'][(0,0)]
            if [p['pdg_code'] for p in candidate['particles']] == [21, 21, 21, 21]:
                input_4g = candidate
                coupling_4g = candidate['couplings'][(0,0)]
            if [p['pdg_code'] for p in candidate['particles']] == [1, 1, 23]:
                input_ddz = candidate
                coupling_ddz_1 = candidate['couplings'][(0,0)]
                coupling_ddz_2 = candidate['couplings'][(0,1)]
            if [p['pdg_code'] for p in candidate['particles']] == [11, 11, 23]:
                input_eez = candidate
                coupling_eez_1 = candidate['couplings'][(0,0)]            
                coupling_eez_2 = candidate['couplings'][(0,1)]
        
        #security                                      
        found_4g = 0  
        found_bbh = 0 
        for dep,data in self.model['couplings'].items():
            for param in data:
                if param.name == coupling_4g: found_4g +=1
                elif param.name == coupling_bbh: found_bbh +=1
        self.assertTrue(found_bbh>0)
        self.assertTrue(found_4g>0)
        
        # make the real test
        self.model.locate_coupling()
        result = self.model.remove_interactions([coupling_bbh, coupling_4g])
        self.assertFalse(input_bbh in self.model['interactions'])
        self.assertFalse(input_4g in self.model['interactions'])
        
    
        # Now test case where some of them are deleted and some not
        if coupling_ddz_1 != coupling_eez_1:
            coupling_eez_1, coupling_eez_2 = coupling_eez_2, coupling_eez_1
        assert coupling_ddz_1 == coupling_eez_1
        
        result = self.model.remove_interactions([coupling_ddz_1, coupling_ddz_2])
        self.assertTrue(coupling_eez_2 in input_eez['couplings'].values())
        self.assertFalse(coupling_eez_1 in input_eez['couplings'].values())
        self.assertFalse(coupling_ddz_1 in input_ddz['couplings'].values())
        self.assertFalse(coupling_ddz_2 in input_ddz['couplings'].values())

    def test_put_parameters_to_zero(self):
        """check that we remove parameters correctly"""
        
        part_t = self.model.get_particle(6)
        # Check that we remove a mass correctly
        self.assertEqual(part_t['mass'], 'MT')
        self.model.fix_parameter_values(['MT'],[])
        self.assertEqual(part_t['mass'], 'ZERO')
        for dep,data in self.model['parameters'].items():
            for param in data:
                self.assertNotEqual(param.name, 'MT')
        
        for particle in self.model['particles']:
            self.assertNotEqual(particle['mass'], 'MT')
                    
        for pdg, particle in self.model['particle_dict'].items():
            self.assertNotEqual(particle['mass'], 'MT')
        
        # Check that we remove a width correctly
        self.assertEqual(part_t['width'], 'WT')
        self.model.fix_parameter_values(['WT'],[])
        self.assertEqual(part_t['width'], 'ZERO')
        for dep,data in self.model['parameters'].items():
            for param in data:
                self.assertNotEqual(param.name, 'WT')

        for pdg, particle in self.model['particle_dict'].items():
            self.assertNotEqual(particle['width'], 'WT')       
             
        # Check that we can remove correctly other external parameter
        self.model.fix_parameter_values(['ymb','yb'],[])
        for dep,data in self.model['parameters'].items():
            for param in data:
                self.assertFalse(param.name in  ['ymb'])
                if param.name == 'yb':
                    param.expr == 'ZERO'
                    
    def test_get_new_coupling_name(self):
        """ test that the static function get_new_coupling_name
            behaves as expected
        """
        
        # reject wrong input
        self.assertRaises(AssertionError, import_ufo.RestrictModel.get_new_coupling_name,
                         '','','',0)
        self.assertRaises(AssertionError, import_ufo.RestrictModel.get_new_coupling_name,
                         '','','',2.)
        self.assertRaises(AssertionError, import_ufo.RestrictModel.get_new_coupling_name,
                         '','1','2',1)      
        self.assertRaises(AssertionError, import_ufo.RestrictModel.get_new_coupling_name,
                         '',1,1,1)         
        self.assertRaises(AssertionError, import_ufo.RestrictModel.get_new_coupling_name,
                         1,'1','1',1)
        
        # real test
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        'GC1', 'GC2', 'GC2', 1), 'GC1')        
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        'GC1', 'GC2', '-GC2', 1), '-GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        'GC1', '-GC2', 'GC2', 1), '-GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        'GC1', '-GC2', '-GC2', 1), 'GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        '-GC1', 'GC2', 'GC2', 1), '-GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        '-GC1', 'GC2', '-GC2', 1), 'GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        '-GC1', '-GC2', 'GC2', 1), 'GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        '-GC1', '-GC2', '-GC2', 1), '-GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        'GC1', 'GC2', 'GC2', -1), '-GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        'GC1', 'GC2', '-GC2', -1), 'GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        'GC1', '-GC2', 'GC2', -1), 'GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        'GC1', '-GC2', '-GC2', -1), '-GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        '-GC1', 'GC2', 'GC2', -1), 'GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        '-GC1', 'GC2', '-GC2', -1), '-GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        '-GC1', '-GC2', 'GC2', -1), '-GC1') 
        self.assertEqual(import_ufo.RestrictModel.get_new_coupling_name(\
                        '-GC1', '-GC2', '-GC2', -1), 'GC1')
                   
    def test_restrict_from_a_param_card(self):
        """ check the full restriction chain in one case b b~ h """
        
        for candidate in self.model['interactions']:
            if [p['pdg_code'] for p in candidate['particles']] == [5, 5, 25]:
                interaction = candidate
                coupling = interaction['couplings'][(0,0)]
    

        self.model.restrict_model(self.restrict_file)

        # check remove interactions
        self.assertFalse(interaction in self.model['interactions'])
        
        # check remove parameters
        for dep,data in self.model['parameters'].items():
            for param in data:
                self.assertFalse(param.name in  ['yb','ymb','MB','WT'])

        # check remove couplings
        for dep,data in self.model['couplings'].items():
            for param in data:
                self.assertFalse(param.name in  [coupling])

        # check masses
        part_b = self.model.get_particle(5)
        part_t = self.model.get_particle(6)
        self.assertEqual(part_b['mass'], 'ZERO')
        self.assertEqual(part_t['width'], 'ZERO')
                
        # check identical masses
        keeped, rejected = None, None 
        for param in self.model['parameters'][('external',)]:
            if param.name == 'MH':
                self.assertEqual(keeped, None)
                keeped, rejected = 'MH','MZ'
            elif param.name == 'MZ':
                self.assertEqual(keeped, None)
                keeped, rejected = 'MZ','MH'
                
        self.assertNotEqual(keeped, None)
        
        found = 0
        for param in self.model['parameters'][()]:
            self.assertNotEqual(param.name, keeped)
            if param.name == rejected:
                found +=1
        self.assertEqual(found, 1)
       
class TestBenchmarkModel(unittest.TestCase):
    """Test class for the RestrictModel object"""

    def setUp(self):
        """Set up decay model"""
        #Read the full SM
        sm_path = import_ufo.find_ufo_path('sm')
        self.base_model = import_ufo.import_full_model(sm_path)
        model = copy.deepcopy(self.base_model)
        self.model = import_ufo.RestrictModel(model)
        self.restrict_file = os.path.join(_file_path, os.path.pardir,
                                     'input_files', 'restrict_sm.dat')
        
        
    def test_use_as_benchmark(self):
        """check that the value inside the restrict card overwritte the default
        parameter such that this option can be use for benchmark point"""
        
        params_ext = self.model['parameters'][('external',)]
        value = {}
        [value.__setitem__(data.name, data.value) for data in params_ext] 
        self.model.restrict_model(self.restrict_file)
        #use the UFO -> MG4 converter class
        
        params_ext = self.model['parameters'][('external',)]
        value2 = {}
        [value2.__setitem__(data.name, data.value) for data in params_ext] 
        
        self.assertNotEqual(value['WW'], value2['WW'])
                
    def test_model_name(self):
        """ test that the model name is correctly set """
        self.assertEqual(self.base_model["name"], "sm")
        model = import_ufo.import_model('sm-full') 
        self.assertEqual(model["name"], "sm-full")
        model = import_ufo.import_model('sm-no_b_mass') 
        self.assertEqual(model["name"], "sm-no_b_mass")        

